@@ -1,12 +1,13 @@
 /*
  * AKFAvatar SDL backend
- * Copyright (c) 2007,2008,2009,2010,2011,2012 Andreas K. Foerster <info@akfoerster.de>
+ * Copyright (c) 2007,2008,2009,2010,2011,2012,2013
+ * Andreas K. Foerster <info@akfoerster.de>
  *
  * required standards: C99
  *
  * other software:
  * required:
- *  SDL1.2.11 or later (but not 1.3!)
+ *  SDL1.2.11 or later
  * optional (deprecated):
  *  SDL_image1.2
  *
@@ -29,9 +30,6 @@
 #define _ISOC99_SOURCE
 #define _XOPEN_SOURCE 600
 
-// undefine to deactivate imageloaders including SDL_Image
-#define IMAGELOADERS
-
 // don't make functions deprecated for this file
 #define _AVT_USE_DEPRECATED
 
@@ -46,8 +44,20 @@
 
 // include images
 #include "akfavatar.xpm"
-#include "mpointer.xbm"
-#include "mpointer_mask.xbm"
+#include "mfinger.xbm"
+#include "mfinger_mask.xbm"
+
+
+#if SDL_VERSION_ATLEAST(1, 3, 0)
+#define SDL2
+#endif
+
+// FIXME: just for porting
+#ifndef SDL2
+// undefine to deactivate imageloaders including SDL_Image
+#define IMAGELOADERS
+#endif
+
 
 #if defined(LINK_SDL_IMAGE) and defined(IMAGELOADERS)
 #  include "SDL_image.h"
@@ -58,6 +68,7 @@
 #define AVT_TIMEOUT 1
 #define AVT_PUSH_KEY 2
 
+#ifndef SDL2
 // only defined in later SDL versions
 #ifndef SDL_BUTTON_WHEELUP
 #  define SDL_BUTTON_WHEELUP 4
@@ -66,13 +77,24 @@
 #ifndef SDL_BUTTON_WHEELDOWN
 #  define SDL_BUTTON_WHEELDOWN 5
 #endif
+#endif // SDL1
 
+#ifdef SDL2
+static SDL_Window *sdl_window;
+static SDL_Renderer *sdl_renderer;
+static SDL_Texture *sdl_screen;
+static const struct avt_charenc *utf8;
+#else // SDL-1.2
 static SDL_Surface *sdl_screen;
-static short int mode;		// whether fullscreen or window or ...
-static SDL_Cursor *mpointer;
-static int windowmode_width, windowmode_height;
+#endif
+
+static SDL_Cursor *mouse_finger;
+
 static Uint32 screenflags;	// flags for the screen
 
+static short int mode;		// whether fullscreen or window or ...
+static int windowmode_width, windowmode_height;
+
 static bool reserve_single_keys;
 static avt_char pointer_button_key;	// key simulated for mouse button 1-3
 static avt_char pointer_motion_key;	// key simulated be pointer motion
@@ -83,6 +105,56 @@ static void avt_analyze_event (SDL_Event * event);
 //-----------------------------------------------------------------------------
 
 
+#ifdef SDL2
+
+// this shall be the only function to update the window/screen
+static void
+update_area_sdl (avt_graphic * screen, int x, int y, int width, int height)
+{
+  SDL_Rect rect;
+  int screen_width, screen_height;
+
+  screen_width = screen->width;
+  screen_height = screen->height;
+
+  if (x < 0)
+    {
+      width -= (-x);
+      x = 0;
+    }
+
+  if (x + width > screen_width)
+    width = screen_width - x;
+
+  if (y < 0)
+    {
+      height -= (-y);
+      y = 0;
+    }
+
+  if (y + height > screen_height)
+    height = screen_height - y;
+
+  if (width <= 0 or height <= 0 or x > screen_width or y > screen_height)
+    return;
+
+  rect.x = x;
+  rect.y = y;
+  rect.w = width;
+  rect.h = height;
+
+  SDL_UpdateTexture (sdl_screen, &rect,
+		     screen->pixels + (y * screen_width) + x,
+		     MINIMALWIDTH * sizeof (avt_color));
+
+  SDL_RenderClear (sdl_renderer);
+  SDL_RenderCopy (sdl_renderer, sdl_screen, NULL, NULL);
+
+  SDL_RenderPresent (sdl_renderer);
+}
+
+#else // SDL-1.2
+
 // this shall be the only function to update the window/screen
 static void
 update_area_sdl (avt_graphic * screen, int x, int y, int width, int height)
@@ -94,6 +166,7 @@ update_area_sdl (avt_graphic * screen, int x, int y, int width, int height)
   SDL_UpdateRect (sdl_screen, x, y, width, height);
 }
 
+#endif // SDL-1.2
 
 #ifdef IMAGELOADERS
 
@@ -189,6 +262,46 @@ load_image_done (void)
 
 // sdl image loaders
 
+#ifdef SDL2
+
+// import an SDL_Surface into the internal format
+static inline avt_graphic *
+avt_import_sdl_surface (SDL_Surface * s)
+{
+  Uint32 flags;
+  avt_graphic *gr;
+  SDL_Surface *d;
+  SDL_PixelFormat format = {
+    SDL_PIXELFORMAT_ARGB8888,
+    NULL,
+    CHAR_BIT * sizeof (avt_color), sizeof (avt_color),
+    {0, 0}
+    ,
+    0x00FF0000, 0x0000FF00, 0x000000FF, 0,
+    0, 0, 0, 0,
+    16, 8, 0, 0,
+    0, NULL
+  };
+
+  flags = 0;			// unused
+
+  // convert into the internally used pixel format
+  d = SDL_ConvertSurface (s, &format, flags);
+
+  gr = avt_new_graphic (d->w, d->h);
+  if (gr)
+    {
+      gr->transparent = (SDL_GetColorKey (s, &gr->color_key) == 0);
+      memcpy (gr->pixels, d->pixels, d->w * d->h * sizeof (avt_color));
+    }
+
+  SDL_FreeSurface (d);
+  // s is freed by the caller
+
+  return gr;
+}
+
+#else // SDL-1.2
 
 // import an SDL_Surface into the internal format
 static inline avt_graphic *
@@ -224,6 +337,7 @@ avt_import_sdl_surface (SDL_Surface * s)
 
   return gr;
 }
+#endif // SDL-1.2
 
 static inline avt_graphic *
 avt_load_image_rw (SDL_RWops * RW)
@@ -276,6 +390,86 @@ load_image_memory_sdl (void *data, size_t size)
 #define load_image_done(void)	// empty
 #endif // not IMAGELOADERS
 
+
+#ifdef SDL2
+
+extern void
+avt_toggle_fullscreen (void)
+{
+  if (not sdl_screen)
+    return;
+
+  // toggle bit for fullscreenmode
+  screenflags = screenflags xor SDL_WINDOW_FULLSCREEN_DESKTOP;
+
+  SDL_SetWindowFullscreen (sdl_window, screenflags);
+
+  if ((screenflags bitand SDL_WINDOW_FULLSCREEN_DESKTOP) != 0)
+    mode = AVT_FULLSCREEN;
+  else
+    mode = AVT_WINDOW;
+}
+
+extern void
+avt_switch_mode (int new_mode)
+{
+  if (sdl_screen and new_mode != mode)
+    {
+      mode = new_mode;
+      switch (mode)
+	{
+	case AVT_FULLSCREENNOSWITCH:
+	case AVT_FULLSCREEN:
+	  if ((screenflags bitand SDL_WINDOW_FULLSCREEN_DESKTOP) == 0)
+	    {
+	      screenflags = screenflags bitor SDL_WINDOW_FULLSCREEN_DESKTOP;
+	      SDL_SetWindowFullscreen (sdl_window, screenflags);
+	    }
+	  break;
+
+	case AVT_WINDOW:
+	  if ((screenflags bitand SDL_WINDOW_FULLSCREEN_DESKTOP) != 0)
+	    {
+	      screenflags =
+		screenflags bitand compl (SDL_WINDOW_FULLSCREEN_DESKTOP);
+	      SDL_SetWindowFullscreen (sdl_window, screenflags);
+	    }
+	  break;
+	}
+    }
+}
+
+// add text (UTF-8) into keyboard buffer
+static inline void
+avt_add_text (const char *text)
+{
+  while (*text)
+    {
+      avt_char ch;
+      size_t s = utf8->decode (utf8, &ch, text);
+      avt_add_key (ch);
+      text += s;
+    }
+}
+
+// add text from clipboard to key buffer
+// up to 512 characters - rest is ignored
+static inline void
+avt_add_clipboard (void)
+{
+  char *text;
+
+  text = SDL_GetClipboardText ();
+
+  if (text)
+    {
+      avt_add_text (text);
+      SDL_free (text);
+    }
+}
+
+#else // SDL-1.2
+
 static void
 resize_sdl (avt_graphic * screen, int width, int height)
 {
@@ -356,10 +550,12 @@ avt_switch_mode (int new_mode)
     }
 }
 
+#endif // SDL-1.2
+
 static inline void
-avt_analyze_key (SDL_keysym key)
+avt_analyze_key (Sint32 keycode, Uint16 mod)
 {
-  switch (key.sym)
+  switch (keycode)
     {
     case SDLK_PAUSE:
       avt_pause ();
@@ -373,10 +569,8 @@ avt_analyze_key (SDL_keysym key)
       break;
 
     case SDLK_q:
-      if (key.mod & KMOD_LALT)
+      if (mod & KMOD_LALT)
 	_avt_STATUS = AVT_QUIT;
-      else if (key.unicode)
-	avt_add_key (key.unicode);
       break;
 
     case SDLK_F11:
@@ -387,99 +581,174 @@ avt_analyze_key (SDL_keysym key)
       break;
 
     case SDLK_RETURN:
-      if (key.mod & KMOD_LALT)
+      if (mod & KMOD_LALT)
 	avt_toggle_fullscreen ();
       else
 	avt_add_key (AVT_KEY_ENTER);
       break;
 
+    case SDLK_TAB:
+      avt_add_key (0x0009);
+      break;
+
     case SDLK_f:
-      if ((key.mod & KMOD_CTRL) and (key.mod & KMOD_LALT))
+      if ((mod & KMOD_CTRL) and (mod & KMOD_LALT))
 	avt_toggle_fullscreen ();
-      else if (key.unicode)
-	avt_add_key (key.unicode);
       break;
 
     case SDLK_UP:
-    case SDLK_KP8:
-      if (key.unicode)
-	avt_add_key (key.unicode);
-      else
-	avt_add_key (AVT_KEY_UP);
+      avt_add_key (AVT_KEY_UP);
       break;
 
     case SDLK_DOWN:
-    case SDLK_KP2:
-      if (key.unicode)
-	avt_add_key (key.unicode);
-      else
-	avt_add_key (AVT_KEY_DOWN);
+      avt_add_key (AVT_KEY_DOWN);
       break;
 
     case SDLK_RIGHT:
-    case SDLK_KP6:
-      if (key.unicode)
-	avt_add_key (key.unicode);
-      else
-	avt_add_key (AVT_KEY_RIGHT);
+      avt_add_key (AVT_KEY_RIGHT);
       break;
 
     case SDLK_LEFT:
-    case SDLK_KP4:
-      if (key.unicode)
-	avt_add_key (key.unicode);
-      else
-	avt_add_key (AVT_KEY_LEFT);
+      avt_add_key (AVT_KEY_LEFT);
       break;
 
     case SDLK_INSERT:
-    case SDLK_KP0:
-      if (key.unicode)
-	avt_add_key (key.unicode);
-      else
-	avt_add_key (AVT_KEY_INSERT);
+      avt_add_key (AVT_KEY_INSERT);
       break;
 
     case SDLK_DELETE:
-    case SDLK_KP_PERIOD:
-      if (key.unicode)
-	avt_add_key (key.unicode);
-      else
-	avt_add_key (AVT_KEY_DELETE);
+      avt_add_key (AVT_KEY_DELETE);
       break;
 
     case SDLK_HOME:
-    case SDLK_KP7:
-      if (key.unicode)
-	avt_add_key (key.unicode);
-      else
-	avt_add_key (AVT_KEY_HOME);
+      avt_add_key (AVT_KEY_HOME);
       break;
 
     case SDLK_END:
-    case SDLK_KP1:
-      if (key.unicode)
-	avt_add_key (key.unicode);
-      else
-	avt_add_key (AVT_KEY_END);
+      avt_add_key (AVT_KEY_END);
       break;
 
     case SDLK_PAGEUP:
-    case SDLK_KP9:
-      if (key.unicode)
-	avt_add_key (key.unicode);
-      else
-	avt_add_key (AVT_KEY_PAGEUP);
+      avt_add_key (AVT_KEY_PAGEUP);
       break;
 
     case SDLK_PAGEDOWN:
-    case SDLK_KP3:
-      if (key.unicode)
-	avt_add_key (key.unicode);
-      else
+      avt_add_key (AVT_KEY_PAGEDOWN);
+      break;
+
+    case SDLK_KP_ENTER:
+      avt_add_key (AVT_KEY_ENTER);
+      break;
+
+#ifdef SDL2
+
+      // alt + v  (get clipboard text)
+    case SDLK_v:
+      if (mod & KMOD_LALT)
+	avt_add_clipboard ();
+      break;
+
+      // Checking for NumLock is unreliable!
+      // So I check if the key caused a text-input event.
+
+    case SDLK_KP_0:
+      if (not SDL_HasEvent (SDL_TEXTINPUT))
+	avt_add_key (AVT_KEY_INSERT);
+      break;
+
+    case SDLK_KP_1:
+      if (not SDL_HasEvent (SDL_TEXTINPUT))
+	avt_add_key (AVT_KEY_END);
+      break;
+
+    case SDLK_KP_2:
+      if (not SDL_HasEvent (SDL_TEXTINPUT))
+	avt_add_key (AVT_KEY_DOWN);
+      break;
+
+    case SDLK_KP_3:
+      if (not SDL_HasEvent (SDL_TEXTINPUT))
 	avt_add_key (AVT_KEY_PAGEDOWN);
       break;
 
+    case SDLK_KP_4:
+      if (not SDL_HasEvent (SDL_TEXTINPUT))
+	avt_add_key (AVT_KEY_LEFT);
+      break;
+
+    case SDLK_KP_6:
+      if (not SDL_HasEvent (SDL_TEXTINPUT))
+	avt_add_key (AVT_KEY_RIGHT);
+      break;
+
+    case SDLK_KP_7:
+      if (not SDL_HasEvent (SDL_TEXTINPUT))
+	avt_add_key (AVT_KEY_HOME);
+      break;
+
+    case SDLK_KP_8:
+      if (not SDL_HasEvent (SDL_TEXTINPUT))
+	avt_add_key (AVT_KEY_UP);
+      break;
+
+    case SDLK_KP_9:
+      if (not SDL_HasEvent (SDL_TEXTINPUT))
+	avt_add_key (AVT_KEY_PAGEUP);
+      break;
+
+    case SDLK_KP_PERIOD:
+      if (not SDL_HasEvent (SDL_TEXTINPUT))
+	avt_add_key (AVT_KEY_DELETE);
+      break;
+
+#else // SDL-1.2
+
+    case SDLK_KP0:
+      avt_add_key (AVT_KEY_INSERT);
+      break;
+
+    case SDLK_KP1:
+      avt_add_key (AVT_KEY_END);
+      break;
+
+    case SDLK_KP2:
+      avt_add_key (AVT_KEY_DOWN);
+      break;
+
+    case SDLK_KP3:
+      avt_add_key (AVT_KEY_PAGEDOWN);
+      break;
+
+    case SDLK_KP4:
+      avt_add_key (AVT_KEY_LEFT);
+      break;
+
+    case SDLK_KP6:
+      avt_add_key (AVT_KEY_RIGHT);
+      break;
+
+    case SDLK_KP7:
+      avt_add_key (AVT_KEY_HOME);
+      break;
+
+    case SDLK_KP8:
+      avt_add_key (AVT_KEY_UP);
+      break;
+
+    case SDLK_KP9:
+      avt_add_key (AVT_KEY_PAGEUP);
+      break;
+
+    case SDLK_KP_PERIOD:
+      avt_add_key (AVT_KEY_DELETE);
+      break;
+
+    case SDLK_EURO:
+      avt_add_key (0x20AC);
+      break;
+
+#endif
+
     case SDLK_BACKSPACE:
       avt_add_key (AVT_KEY_BACKSPACE);
       break;
@@ -492,10 +761,6 @@ avt_analyze_key (SDL_keysym key)
       avt_add_key (AVT_KEY_MENU);
       break;
 
-    case SDLK_EURO:
-      avt_add_key (0x20AC);
-      break;
-
     case SDLK_F1:
     case SDLK_F2:
     case SDLK_F3:
@@ -507,17 +772,16 @@ avt_analyze_key (SDL_keysym key)
     case SDLK_F9:
     case SDLK_F10:
     case SDLK_F12:
-    case SDLK_F13:
-    case SDLK_F14:
-    case SDLK_F15:
-      avt_add_key (AVT_KEY_F1 + (key.sym - SDLK_F1));
-      break;
-
-    default:
-      if (key.unicode)
-	avt_add_key (key.unicode);
+      avt_add_key (AVT_KEY_F1 + (keycode - SDLK_F1));
       break;
     }				// switch (key.sym)
+
+#ifdef SDL2
+  // Ctrl + letter
+  if ((mod & KMOD_CTRL) and (mod & KMOD_LALT) == 0
+      and (keycode >= SDLK_a and keycode <= SDLK_z))
+    avt_add_key (keycode - (SDLK_a - 1));
+#endif
 }
 
 static void
@@ -529,12 +793,9 @@ avt_analyze_event (SDL_Event * event)
       _avt_STATUS = AVT_QUIT;
       break;
 
-    case SDL_VIDEORESIZE:
-      avt_resize (event->resize.w, event->resize.h);
-      break;
-
     case SDL_MOUSEBUTTONDOWN:
-      if (event->button.button <= 3)
+      if (SDL_BUTTON_LEFT == event->button.button
+	  or SDL_BUTTON_RIGHT == event->button.button)
 	{
 	  if (not avt_check_buttons (event->button.x, event->button.y))
 	    {
@@ -542,10 +803,15 @@ avt_analyze_event (SDL_Event * event)
 		avt_add_key (pointer_button_key);
 	    }
 	}
+#ifdef SDL2
+      else if (SDL_BUTTON_MIDDLE == event->button.button)
+	avt_add_clipboard ();
+#else // SDL-1.2
       else if (SDL_BUTTON_WHEELDOWN == event->button.button)
 	avt_add_key (AVT_KEY_DOWN);
       else if (SDL_BUTTON_WHEELUP == event->button.button)
 	avt_add_key (AVT_KEY_UP);
+#endif
       break;
 
     case SDL_MOUSEMOTION:
@@ -553,9 +819,59 @@ avt_analyze_event (SDL_Event * event)
 	avt_add_key (pointer_motion_key);
       break;
 
-    case SDL_KEYDOWN:
-      avt_analyze_key (event->key.keysym);
+#ifdef SDL2
+    case SDL_WINDOWEVENT:
+      switch (event->window.event)
+	{
+	case SDL_WINDOWEVENT_SHOWN:
+	case SDL_WINDOWEVENT_EXPOSED:
+	case SDL_WINDOWEVENT_SIZE_CHANGED:
+	  avt_update_all ();	// redraw
+	  break;
+	}
       break;
+
+    case SDL_KEYDOWN:
+      avt_analyze_key (event->key.keysym.sym, event->key.keysym.mod);
+      break;
+
+    case SDL_TEXTINPUT:
+      avt_add_text (event->text.text);
+      break;
+
+    case SDL_MOUSEWHEEL:
+      if (event->wheel.y < 0)
+	for (int i = -(event->wheel.y); i > 0; --i)
+	  avt_add_key (AVT_KEY_DOWN);
+      else if (event->wheel.y > 0)
+	for (int i = event->wheel.y; i > 0; --i)
+	  avt_add_key (AVT_KEY_UP);
+
+      if (event->wheel.x < 0)
+	for (int i = -(event->wheel.x); i > 0; --i)
+	  avt_add_key (AVT_KEY_LEFT);
+      else if (event->wheel.x > 0)
+	for (int i = event->wheel.x; i > 0; --i)
+	  avt_add_key (AVT_KEY_RIGHT);
+      break;
+
+#else // SDL-1.2
+
+    case SDL_VIDEORESIZE:
+      avt_resize (event->resize.w, event->resize.h);
+      break;
+
+#define ESC  (27)
+
+    case SDL_KEYDOWN:
+      if (event->key.keysym.unicode
+	  and event->key.keysym.unicode != ESC
+	  and not (event->key.keysym.mod bitand KMOD_LALT))
+	avt_add_key (event->key.keysym.unicode);
+      else
+	avt_analyze_key (event->key.keysym.sym, event->key.keysym.mod);
+      break;
+#endif
     }				// switch (event->type)
 }
 
@@ -736,16 +1052,30 @@ avt_set_pointer_buttons_key (avt_char key)
 extern void
 avt_get_pointer_position (int *x, int *y)
 {
-  if (sdl_screen)
-    SDL_GetMouseState (x, y);
-  else
-    {
-      if (x)
-	*x = 0;
+  int mx, my;
 
-      if (y)
-	*y = 0;
+  mx = my = 0;
+
+  if (sdl_screen)
+    {
+#ifdef SDL2
+      // compute scale
+      int wx, wy;
+      SDL_GetWindowSize (sdl_window, &wx, &wy);
+      SDL_GetMouseState (&mx, &my);
+
+      mx = mx * MINIMALWIDTH / wx;
+      my = my * MINIMALHEIGHT / wy;
+#else
+      SDL_GetMouseState (&mx, &my);
+#endif
     }
+
+  if (x)
+    *x = mx;
+
+  if (y)
+    *y = my;
 }
 
 extern void
@@ -769,7 +1099,7 @@ avt_get_mode (void)
 extern char *
 avt_get_error (void)
 {
-  return SDL_GetError ();
+  return (char *) SDL_GetError ();
 }
 
 extern void
@@ -787,11 +1117,13 @@ avt_init_SDL (void)
   // only if not already initialized
   if (SDL_WasInit (SDL_INIT_VIDEO | SDL_INIT_TIMER) == 0)
     {
+#ifndef SDL2
       /* don't try to use the mouse
        * might be needed for the fbcon driver
        * the mouse still works, it is just not required
        */
       SDL_putenv ("SDL_NOMOUSE=1");
+#endif
 
       if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0)
 	_avt_STATUS = AVT_ERROR;
@@ -808,58 +1140,65 @@ quit_sdl (void)
 
   if (sdl_screen)
     {
-      SDL_FreeCursor (mpointer);
-      mpointer = NULL;
+      SDL_FreeCursor (mouse_finger);
+      mouse_finger = NULL;
       SDL_Quit ();
       sdl_screen = NULL;	// it was freed by SDL_Quit
     }
 }
 
+#ifdef SDL2
+static void
+background_color_sdl (avt_color c)
+{
+  SDL_SetRenderDrawColor (sdl_renderer, avt_red (c), avt_green (c),
+			  avt_blue (c), SDL_ALPHA_OPAQUE);
+}
+#endif
+
 extern void
 avt_set_title (const char *title, const char *shortname)
 {
+#ifdef SDL2
+  SDL_SetWindowTitle (sdl_window, title);
+#else
   // assumes UTF-8
   SDL_WM_SetCaption (title, shortname);
+#endif
 }
 
+static inline void
+reverse_bits (unsigned char *bytes, size_t length)
+{
+  while (length--)
+    {
+      register unsigned char b = *bytes;
 
-#define reverse_byte(b) \
-  (((b) & 0x80) >> 7 | \
-   ((b) & 0x40) >> 5 | \
-   ((b) & 0x20) >> 3 | \
-   ((b) & 0x10) >> 1 | \
-   ((b) & 0x08) << 1 | \
-   ((b) & 0x04) << 3 | \
-   ((b) & 0x02) << 5 | \
-   ((b) & 0x01) << 7)
+      *bytes = (b bitand 0x80) >> 7
+	bitor (b bitand 0x40) >> 5
+	bitor (b bitand 0x20) >> 3
+	bitor (b bitand 0x10) >> 1
+	bitor (b bitand 0x08) << 1
+	bitor (b bitand 0x04) << 3
+	bitor (b bitand 0x02) << 5 bitor (b bitand 0x01) << 7;
 
-#define xbm_bytes(img)  (((img##_width+CHAR_BIT-1) / CHAR_BIT) * img##_height)
+      ++bytes;
+    }
+}
 
 static inline void
-avt_set_mouse_pointer (void)
+avt_prepare_mouse_cursors (void)
 {
-  unsigned char mp[xbm_bytes (mpointer)];
-  unsigned char mp_mask[xbm_bytes (mpointer_mask)];
-
-  // we need the bytes reversed :-(
-
-  for (int i = 0; i < xbm_bytes (mpointer); i++)
+  if (not mouse_finger)
     {
-      register unsigned char b = mpointer_bits[i];
-      mp[i] = reverse_byte (b);
+      // we need the bits reversed :-(
+      reverse_bits (mfinger_bits, sizeof (mfinger_bits));
+      reverse_bits (mfinger_mask_bits, sizeof (mfinger_mask_bits));
+
+      mouse_finger = SDL_CreateCursor (mfinger_bits, mfinger_mask_bits,
+				       mfinger_width, mfinger_height,
+				       mfinger_x_hot, mfinger_y_hot);
     }
-
-  for (int i = 0; i < xbm_bytes (mpointer_mask); i++)
-    {
-      register unsigned char b = mpointer_mask_bits[i];
-      mp_mask[i] = reverse_byte (b);
-    }
-
-  mpointer = SDL_CreateCursor (mp, mp_mask,
-			       mpointer_width, mpointer_height,
-			       mpointer_x_hot, mpointer_y_hot);
-
-  SDL_SetCursor (mpointer);
 }
 
 static inline void
@@ -875,11 +1214,22 @@ avt_set_icon (char **xpm)
 				   gr->width * sizeof (avt_color),
 				   0x00FF0000, 0x0000FF00, 0x000000FF, 0);
 
+#ifdef SDL2
+
+  if (gr->transparent)
+    SDL_SetColorKey (icon, SDL_TRUE, gr->color_key);
+
+  SDL_SetWindowIcon (sdl_window, icon);
+
+#else // SDL-1.2
+
   if (gr->transparent)
     SDL_SetColorKey (icon, SDL_SRCCOLORKEY, gr->color_key);
 
   SDL_WM_SetIcon (icon, NULL);
 
+#endif
+
   SDL_FreeSurface (icon);
   avt_free_graphic (gr);
 }
@@ -912,9 +1262,67 @@ avt_start (const char *title, const char *shortname, int window_mode)
   if (not shortname)
     shortname = title;
 
-  avt_set_title (title, shortname);
+#ifdef SDL2
+  screenflags = SDL_WINDOW_RESIZABLE;
+
+  if (mode >= 1)
+    screenflags |= SDL_WINDOW_FULLSCREEN_DESKTOP;
+
+  sdl_window =
+    SDL_CreateWindow (title, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
+		      MINIMALWIDTH, MINIMALHEIGHT, screenflags);
+
+  if (not sdl_window)
+    {
+      avt_set_error ("error creating window");
+      _avt_STATUS = AVT_ERROR;
+      return _avt_STATUS;
+    }
+
   avt_set_icon (akfavatar_xpm);
 
+  SDL_SetWindowMinimumSize (sdl_window, MINIMALWIDTH, MINIMALHEIGHT);
+
+  sdl_renderer = SDL_CreateRenderer (sdl_window, -1, 0);
+
+  if (not sdl_renderer)
+    {
+      avt_set_error ("error creating renderer");
+      _avt_STATUS = AVT_ERROR;
+      return _avt_STATUS;
+    }
+
+  SDL_SetHint (SDL_HINT_RENDER_SCALE_QUALITY, "linear");
+  SDL_RenderSetLogicalSize (sdl_renderer, MINIMALWIDTH, MINIMALHEIGHT);
+
+  sdl_screen = SDL_CreateTexture (sdl_renderer, SDL_PIXELFORMAT_ARGB8888,
+				  SDL_TEXTUREACCESS_STREAMING,
+				  MINIMALWIDTH, MINIMALHEIGHT);
+
+  if (not sdl_screen)
+    {
+      avt_set_error ("error creating screen texture");
+      _avt_STATUS = AVT_ERROR;
+      return _avt_STATUS;
+    }
+
+  // set up a new graphic to draw to
+  backend = avt_start_common (avt_new_graphic (MINIMALWIDTH, MINIMALHEIGHT));
+
+  background_color_sdl (avt_get_background_color ());
+
+  // size of the window (not to be confused with the variable window)
+  windowmode_width = MINIMALWIDTH;
+  windowmode_height = MINIMALHEIGHT;
+
+  utf8 = avt_utf8 ();
+
+  SDL_StartTextInput ();
+
+#else // SDL-1.2
+
+  avt_set_title (title, shortname);
+
   // Initialize the display
   screenflags = SDL_SWSURFACE | SDL_RESIZABLE;
 
@@ -934,6 +1342,7 @@ avt_start (const char *title, const char *shortname, int window_mode)
     }
 #endif
 
+  avt_set_icon (akfavatar_xpm);
   SDL_ClearError ();
 
   if (mode >= 1)
@@ -969,18 +1378,34 @@ avt_start (const char *title, const char *shortname, int window_mode)
       return _avt_STATUS;
     }
 
+  // needed to get the character of the typed key
+  SDL_EnableUNICODE (1);
+
+  // key repeat mode
+  SDL_EnableKeyRepeat (SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
+
   // set up a graphic with the same pixel data
   backend = avt_start_common (avt_data_to_graphic (sdl_screen->pixels,
 						   sdl_screen->w,
 						   sdl_screen->h));
 
+  // size of the window (not to be confused with the variable window)
+  windowmode_width = sdl_screen->w;
+  windowmode_height = sdl_screen->h;
+
+#endif /* SDL-1.2 */
+
   if (not backend or _avt_STATUS != AVT_NORMAL)
     return _avt_STATUS;
 
   backend->update_area = update_area_sdl;
   backend->quit = quit_sdl;
   backend->wait_key = wait_key_sdl;
+#ifdef SDL2
+  backend->background_color = background_color_sdl;
+#else
   backend->resize = resize_sdl;
+#endif
 
 #ifdef IMAGELOADERS
   backend->graphic_file = load_image_file_sdl;
@@ -988,17 +1413,8 @@ avt_start (const char *title, const char *shortname, int window_mode)
   backend->graphic_memory = load_image_memory_sdl;
 #endif
 
-  // size of the window (not to be confused with the variable window
-  windowmode_width = sdl_screen->w;
-  windowmode_height = sdl_screen->h;
-
-  avt_set_mouse_pointer ();
-
-  // needed to get the character of the typed key
-  SDL_EnableUNICODE (1);
-
-  // key repeat mode
-  SDL_EnableKeyRepeat (SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
+  avt_prepare_mouse_cursors ();
+  SDL_SetCursor (mouse_finger);
 
   // ignore what we don't use
   SDL_EventState (SDL_MOUSEMOTION, SDL_IGNORE);
